package org.integratedmodelling.thinklab.actions

import org.apache.commons.math3.special.Erf
import org.integratedmodelling.api.knowledge.IConcept
import org.integratedmodelling.api.knowledge.IKnowledge
import org.integratedmodelling.api.knowledge.IObservation
import org.integratedmodelling.api.knowledge.IProperty
import org.integratedmodelling.api.modelling.IDirectObservation
import org.integratedmodelling.api.modelling.IEvent
import org.integratedmodelling.api.modelling.IProcess
import org.integratedmodelling.api.modelling.IRelationship
import org.integratedmodelling.api.modelling.IState
import org.integratedmodelling.api.modelling.ISubject
import org.integratedmodelling.api.modelling.scheduling.ITransition
import org.integratedmodelling.api.monitoring.IMonitor
import org.integratedmodelling.api.monitoring.Messages
import org.integratedmodelling.common.configuration.KLAB
import org.integratedmodelling.common.kim.ModelFactory
import org.integratedmodelling.common.vocabulary.ObservableSemantics
import org.integratedmodelling.engine.modelling.runtime.Subject
import org.integratedmodelling.thinklab.script.DirectObservation
import org.integratedmodelling.thinklab.script.EventInfo
import org.integratedmodelling.thinklab.script.ProcessInfo
import org.integratedmodelling.thinklab.script.Scale
import org.integratedmodelling.thinklab.script.Space
import org.integratedmodelling.thinklab.script.State
import org.integratedmodelling.thinklab.script.StateInfo
import org.integratedmodelling.thinklab.script.SubjectInfo
import org.integratedmodelling.thinklab.script.Time
import org.integratedmodelling.thinklab.script.Transition


//static UnitBinding binding = null;

/**
 * The base class for a Thinklab action. No space or time semantics built in.
 * 
 * @author Ferd
 *
 */
abstract class DefaultAction extends Script {

    static inited = false;
    boolean _wrapped = false;
    private Random rgen = new Random();

    //    Binding binding;

    //    UnitBinding binding;
    //
    //    static class UnitBinding extends Binding {
    //
    //        def getVariable(String symbol) {
    //            if (symbol == 'out') return System.out
    //            return Amount.valueOf(1, Unit.valueOf(symbol))
    //        }
    //    }

    DefaultAction() {
        if (!inited) {
            //			Number.metaClass.getProperty = { String symbol ->
            //				Amount.valueOf(delegate, Unit.valueOf(symbol))
            //			}
            //
            //			// define opeartor overloading, as JScience doesn\'t use the same operation names as Groovy
            //			Amount.metaClass.static.valueOf = { Number number, String unit -> Amount.valueOf(number, Unit.valueOf(unit)) }
            //			Amount.metaClass.multiply = { Number factor -> delegate.times(factor) }
            //			Number.metaClass.multiply = { Amount amount -> amount.times(delegate) }
            //			Number.metaClass.div = { Amount amount -> amount.inverse().times(delegate) }
            //			Amount.metaClass.div = { Number factor -> delegate.divide(factor) }
            //			Amount.metaClass.div = { Amount factor -> delegate.divide(factor) }
            //			Amount.metaClass.power = { Number factor -> delegate.pow(factor) }
            //			Amount.metaClass.negative = { -> delegate.opposite() }
            //
            //			// for unit conversions
            //			Amount.metaClass.to = { Amount amount -> delegate.to(amount.unit) }
            //			Amount.metaClass.to = { String unit -> delegate.to(Unit.valueOf(unit)) }

            //            binding = new Binding();
        }
    }

    /**
     * Compiled in before anything happens; will check for context and target and
     * set up any proxies.
     */
    def wrap() {
        if (!_wrapped) {
            _wrapped = true;
            if (binding.hasVariable("_context")) {
                wrapObject(binding.getVariable("_context"), true);
            }
            if (binding.hasVariable("_self")) {
                wrapObject(binding.getVariable("_self"), false);
            }
        }
        if (binding.hasVariable("_transition") && binding.getVariable("_transition") != null) {
            binding.setVariable("now", new Transition(((ITransition)binding.getVariable('_transition'))));
        }
    }

    /**
     * Random function is always available. TODO establish a mechanism to
     * set the seed. 
     * @return
     */
    def random() {
        rgen.nextDouble();
    }

    /**
     * Missing math functions, using Apache math
     * 
     * @param observation
     * @param isContext
     * @return
     */
    def erf(double d) {
        Erf.erf(d);
    }

    /**
     * Missing math functions, using Apache math
     *
     * @param observation
     * @param isContext
     * @return
     */
    def erfc(double d) {
        Erf.erfc(d);
    }

    /**
     * Missing math functions, using Apache math
     *
     * @param observation
     * @param isContext
     * @return
     */
    def erf(double d1, double d2) {
        Erf.erf(d1, d2);
    }

    /**
     * Missing math functions, using Apache math
     *
     * @param observation
     * @param isContext
     * @return
     */
    def erfInv(double d) {
        Erf.erfInv(d);
    }

    /**
     * Missing math functions, using Apache math
     *
     * @param observation
     * @param isContext
     * @return
     */
    def erfcInv(double d) {
        Erf.erfcInv(d);
    }

    def wrapObject(IObservation observation, boolean isContext) {

        // expose states directly as well as within their own contexts. Context is wrapped
        // first, so overriding works properly.
//        if (observation instanceof IDirectObservation) {
//            for (IState state : ((IDirectObservation)observation).getStates()) {
//                binding.setVariable(state.getObservable().getSemantics().getFormalName(), new State(state, binding));
//            }
//        }

        if (observation instanceof Subject) {
            DirectObservation obs = new DirectObservation(observation, binding);
            binding.setVariable(isContext ? "context" : getObservationId(observation), obs);
        }

        if (!isContext) {
            binding.setVariable('scale', new Scale(observation.getScale(), binding));
            binding.setVariable('space', new Space(observation.getScale().getSpace(), binding));
            binding.setVariable('time', new Time(observation.getScale().getTime(), binding));
        }
    }

    String getObservationId(IObservation obs) {
        if (obs instanceof ISubject) {
            return "subject"
        }
        if (obs instanceof IEvent) {
            return "event"
        }
        if (obs instanceof IProcess) {
            return "process"
        }
        if (obs instanceof IRelationship) {
            return "relationship"
        }
        return "nothing";
    }

    def info(String text) {
        Object o = binding.getVariable("_monitor");
        if (o instanceof IMonitor) {
            ((IMonitor)o).info(text, Messages.INFOCLASS_USER_OWN);
        }
    }

    def warn(String text) {
        Object o = binding.getVariable("_monitor");
        if (o instanceof IMonitor) {
            ((IMonitor)o).warn(text);
        }
    }

    def error(String text) {
        Object o = binding.getVariable("_monitor");
        if (o instanceof IMonitor) {
            ((IMonitor)o).error(text);
        }
    }

    /*
     * window into the knowledge manager
     * TODO use a wrapper that caches is() operations - also when
     * passing to script.
     */
    IConcept _getConcept(String string) {
        return KLAB.c(string);
    }

    IProperty _getProperty(String string) {
        return KLAB.p(string);
    }

    IKnowledge _getKnowledge(String string) {
        return KLAB.k(string);
    }

    SubjectInfo subject(IConcept observable) {
        return new SubjectInfo(observable);
    }

    ProcessInfo process(IConcept observable) {
        return new ProcessInfo(observable);
    }

    EventInfo event(IConcept observable) {
        return new EventInfo(observable);
    }

    StateInfo measure(IConcept observable, String unit, Object value) {
        return new StateInfo(new ObservableSemantics(ModelFactory.measureObserver(observable, unit), null, observable.getLocalName()), value)
    }

    StateInfo measure(IConcept observable, String unit) {
        return new StateInfo(new ObservableSemantics(ModelFactory.measureObserver(observable, unit), null, observable.getLocalName()), null)
    }

    StateInfo rank(IConcept observable, Object value) {
        return new StateInfo(new ObservableSemantics(ModelFactory.rankObserver(observable), null, observable.getLocalName()), value)
    }

    StateInfo rank(IConcept observable) {
        return new StateInfo(new ObservableSemantics(ModelFactory.rankObserver(observable), null, observable.getLocalName()), null)
    }

}
