package org.integratedmodelling.thinklab.script

import org.integratedmodelling.api.knowledge.IConcept
import org.integratedmodelling.api.modelling.IDirectObservation
import org.integratedmodelling.api.modelling.INumericObserver
import org.integratedmodelling.api.modelling.IScale
import org.integratedmodelling.api.modelling.IState
import org.integratedmodelling.common.states.States
import org.integratedmodelling.common.vocabulary.NS
import org.integratedmodelling.engine.visualization.VisualizationFactory

class State extends Observation {

    State(IState obs, Binding binding) {
        super(obs, binding);
    }

    String toString() {
        return obs.toString();
    }

    Object asType(Class cls) {

        Number nret = null;

        if (cls == Number) {
            return aggregateNumbers();
        } else if (cls == Double || cls == double) {
            nret = aggregateNumbers();
            return nret ? nret.doubleValue() : null;
        } else if (cls == Long || cls == long) {
            nret = aggregateNumbers();
            return nret ? nret.longValue() : null;
        } else if (cls == Integer || cls == int) {
            nret = aggregateNumbers();
            return nret ? nret.intValue() : null;
        } else if (cls == Float || cls == float) {
            nret = aggregateNumbers();
            return nret ? nret.floatValue() : null;
        } else if (cls == Boolean || cls == boolean) {
            Object zio = aggregateMajority(Boolean);
            return (zio instanceof Boolean) ? (Boolean)zio : false;
        } else if (cls == IConcept) {
            Object zio = aggregateMajority(IConcept);
            return (zio instanceof IConcept) ? (IConcept)zio : false;
        } else if (cls == double[]) {
            return getDoubleData();
        } else if (cls == long[]) {
            //			return (long)aggregate();
        } else if (cls == int[]) {
            //			return (int)aggregate();
        } else if (cls == float[]) {
            //			return (float)aggregate();
        } else if (cls == boolean[]) {
            // TODO
        } else if (cls == IConcept[]) {
            // TODO
        }
        return null;
    }

    double[] getDoubleData() {
        VisualizationFactory.get().getStateDataAsNumbers((IState)obs, 
            Collections.singleton(binding.hasVariable('_transition') ? binding.getVariable('_transition') : null));
    }

    /**
     * Assign to state as a whole using state << value
     * @param value
     * @return
     */
    def leftShift(Object value) {
        States.set(((IState)obs), value, getTransition());
    }

    /**
     * Support for state[n]
     * 
     * @param offset
     * @return
     */
    def getAt(int offset) {
        States.get((IState)obs, offset, getTransition());
    }

    def at(IDirectObservation o) {
        new State(States.getView((IState)obs, o), binding);
    }

    /**
     * Support for state[n] = value
     * 
     * @param offset
     * @param value
     * @return
     */
    def putAt(int offset, Object value) {
        States.set(((IState)obs), value, offset);    
    }
    
    /**
     * Iterable offsets over the current transition: foreach (n : state.offsets) {}
     * 
     * @return
     */
    def getOffsets() {
        def locs = new IScale.Locator[1];
        locs[0] = getTransition();
        return obs.getScale().getIndex(locs);
    }
    
    Number aggregateNumbers() {
        
        double t = Double.NaN;
        if (((IState)obs).getObserver() instanceof INumericObserver) {
            double[] d = VisualizationFactory.get().getStateDataAsNumbers((IState)obs, 
                Collections.singleton(binding.hasVariable('_transition') ? binding.getVariable('_transition') : null));
            int n = 0;
            for (double v in d) {
                if (!Double.isNaN(v)) {
                    if (Double.isNaN(t)) {
                        t = v;
                    } else {
                        t += v;
                    }
                    n ++
                }
            }

            if (n > 0 && !((IState)obs).getObservable().getType().is(NS.CORE_EXTENSIVE_PHYSICAL_PROPERTY)) {
                t = t/(double)n;
            }
        }
        return t;
    }

    Object aggregateMajority(Class cls) {
        Object ret = null;
        return ret;
    }

    def getSum() {
        double t = Double.NaN;
        if (((IState)obs).getObserver() instanceof INumericObserver) {
            double[] d = VisualizationFactory.get().getStateDataAsNumbers((IState)obs, 
                Collections.singleton(binding.hasVariable('_transition') ? binding.getVariable('_transition') : null));
            int n = 0;
            for (double v in d) {
                if (!Double.isNaN(v)) {
                    if (Double.isNaN(t)) {
                        t = v;
                    } else {
                        t += v;
                    }
                    n ++
                }
            }
        }
        return t;
    }

    /*
     * TODO smarten up - compute stats once per transition and store them.
     */
    def getAvg() {
        double t = Double.NaN;
        if (((IState)obs).getObserver() instanceof INumericObserver) {
            double[] d = VisualizationFactory.get().getStateDataAsNumbers((IState)obs, 
                Collections.singleton(binding.hasVariable('_transition') ? binding.getVariable('_transition') : null));
            int n = 0;
            for (double v in d) {
                if (!Double.isNaN(v)) {
                    if (Double.isNaN(t)) {
                        t = v;
                    } else {
                        t += v;
                    }
                    n ++
                }
            }
            
            if (n > 0) {
                t = t/(double)n;
            }
        }
        return t;
    }

    def getMin() {
        double t = Double.NaN;
        if (((IState)obs).getObserver() instanceof INumericObserver) {
            double[] d = VisualizationFactory.get().getStateDataAsNumbers((IState)obs, 
                Collections.singleton(binding.hasVariable('_transition') ? binding.getVariable('_transition') : null));
            int n = 0;
            for (double v in d) {
                if (!Double.isNaN(v)) {
                    if (Double.isNaN(t) || t > v) {
                        t = v;
                    }
                    n ++
                }
            }
        }
        return t;
    }

    def getMax() {
        double t = Double.NaN;
        if (((IState)obs).getObserver() instanceof INumericObserver) {
            double[] d = VisualizationFactory.get().getStateDataAsNumbers((IState)obs, 
                Collections.singleton(binding.hasVariable('_transition') ? binding.getVariable('_transition') : null));
            int n = 0;
            for (double v in d) {
                if (!Double.isNaN(v)) {
                    if (Double.isNaN(t) || t < v) {
                        t = v;
                    }
                    n ++
                }
            }
        }
        return t;
    }
}
